package org.jmockring.webserver.tomcat;

import java.util.concurrent.CountDownLatch;
import javax.servlet.ServletException;

import org.apache.catalina.Container;
import org.apache.catalina.Context;
import org.apache.catalina.Lifecycle;
import org.apache.catalina.LifecycleEvent;
import org.apache.catalina.LifecycleException;
import org.apache.catalina.LifecycleListener;
import org.apache.catalina.startup.Tomcat;
import org.apache.catalina.util.ServerInfo;
import org.apache.commons.io.FilenameUtils;
import org.jmockring.configuration.ConfigurationConstants;
import org.jmockring.configuration.ServerConfiguration;
import org.jmockring.configuration.ServerExecutionRegistry;
import org.jmockring.configuration.WebAppContextConfiguration;
import org.jmockring.webserver.WebServer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Pavel Lechev <pavel@jmockring.org>
 * @since 12/07/13
 */
public class TomcatWebServer implements WebServer, LifecycleListener {


    private static final Logger log = LoggerFactory.getLogger(TomcatWebServer.class);

    private Tomcat tomcat;

    private ServerConfiguration configuration;

    private CountDownLatch blockingLatch;

    private CountDownLatch latchOnAfterStart;

    @Override
    public void initialise(ServerConfiguration configuration) {

        this.configuration = configuration;
        String location = configuration.getServerConfig().testClass().getProtectionDomain().getCodeSource().getLocation().getPath();

        tomcat = new Tomcat();
        tomcat.setBaseDir(FilenameUtils.normalize(location + "/../tomcat-work"));
        tomcat.setHostname(configuration.getServerConfig().host());
        tomcat.setPort(configuration.getPort());
        tomcat.setSilent(false);
        tomcat.getEngine().setDefaultHost(configuration.getServerConfig().host());
        tomcat.getServer().addLifecycleListener(this);

        latchOnAfterStart = new CountDownLatch(1);
        blockingLatch = new CountDownLatch(1);

        /* There needs to be a symlink to the current dir named 'webapps' */
        try {
            addWebAppContexts(configuration, location);
            tomcat.init();
            addListeners();
        } catch (LifecycleException e) {
            log.error("LOG00140: Failed to initialise Tomcat ", e);
        } catch (ServletException e) {
            log.error("LOG00180: Failed to add web application to Tomcat", e);
        }
    }


    private void addListeners() {
        for (Container container : tomcat.getHost().findChildren()) {
            ((Context) container).addServletContainerInitializer(new ServerExecutionRegistry(), null);
            log.info("LOG00230: Adding registry as listener ...");
        }
    }


    /**
     *
     * @param configuration
     * @param location
     *
     * @throws ServletException
     */
    private void addWebAppContexts(ServerConfiguration configuration, String location) throws ServletException {
        for (WebAppContextConfiguration webAppContext : configuration.getWebAppContexts()) {
            String webAppPath = FilenameUtils.normalize(location + webAppContext.getConfig().webApp());
            Context webContext = tomcat.addWebapp(webAppContext.getConfig().contextPath(), webAppPath);
            webContext.getServletContext().setAttribute(ConfigurationConstants.SERVER_CONFIGURATION_KEY, configuration);
            webContext.getServletContext().setAttribute(ConfigurationConstants.CONTEXT_CONFIGURATION_KEY, webAppContext);
            log.info("LOG00150: Adding web application for context [{}]", webContext.getServletContext());
        }
    }

    @Override
    public void start() {
        try {
            log.info("LOG00170: Starting '{}' on port {} ... ", getName(), getPort());
            tomcat.start();
            blockingLatch.await();  // block here until the #shutdown() is called.
        } catch (LifecycleException e) {
            log.error("LOG00120: Failed to start Tomcat", e);
            throw new RuntimeException(e);
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void shutdown() {
        try {
            log.info("LOG00160: Stopping '{}' on port {} ...", getName(), getPort());
            tomcat.stop();
            blockingLatch.countDown();  // release the block.
        } catch (LifecycleException e) {
            log.error("LOG00120: Failed to stop Tomcat", e);
            throw new RuntimeException(e);
        }
    }

    @Override
    public int getPort() {
        return configuration.getPort();
    }

    @Override
    public String getName() {
        return ServerInfo.getServerInfo();
    }

    @Override
    public void waitForInitialisation() {
        try {
            log.info("LOG00260: Waiting for '{}' to initialise ... ", getName());
            latchOnAfterStart.await();
        } catch (InterruptedException e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void lifecycleEvent(LifecycleEvent event) {
        if (event.getType().equals(Lifecycle.AFTER_START_EVENT)) {
            log.info("LOG00250: [TOMCAT EVENT RECEIVED {}]: unlatching initialisation ... ", event.getType());
            latchOnAfterStart.countDown();
        }
    }
}
